@@ -3,19 +3,27 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
+import argparse
 import sys
 
 from collections import OrderedDict
+from typing import (
+    Container,
+    Dict,
+    List,
+    Optional,
+)
 
 from mozphab.conduit import conduit
 from mozphab.exceptions import Error
 from mozphab.logger import logger
 from mozphab.helpers import augment_commits_from_body, prompt
+from mozphab.repository import Repository
 from mozphab.spinner import wait_message
 from mozphab.telemetry import telemetry
 
 
-def to_llist(revisions):
+def to_llist(revisions: List[str]) -> Dict[str, Optional[str]]:
     """Converts a list to a linked list.
 
     Parameters:
@@ -31,44 +39,51 @@ def to_llist(revisions):
     return llist
 
 
-def walk_llist(llist, allow_multiple_heads=False):
+def walk_llist(
+    llist: Dict[str, Optional[str]], allow_multiple_heads: bool = False
+) -> List[str]:
     """Parse the llist for multiple heads and return a unique list of elements.
 
     Parameters:
         llist (dict): A linked list. {A_id: B_id, B_id: None}
         allow_multiple_heads (bool): Fail if multiple heads found.
     """
-    referenced_children = [r for r in llist.values() if r]
+    referenced_children = [child for child in llist.values() if child]
 
     # Find head
-    node = None
-    for n in sorted(llist.keys()):
-        if n not in referenced_children:
-            if node:
+    head = None
+    for parent in sorted(llist.keys()):
+        if parent not in referenced_children:
+            if head:
                 if not allow_multiple_heads:
                     raise Error("Multiple heads found.")
                 break
-            node = n
-    if not node:
+            head = parent
+    if not head:
         raise Error("Failed to find head.")
 
     # Walk list, checking for loops
     nodes = []
-    while node:
-        nodes.append(node)
-        child = llist.get(node)
+    while head:
+        nodes.append(head)
+        child = llist.get(head)
         if child and child in nodes:
             raise Error("Dependency loop")
 
-        node = child
+        head = child
 
     return nodes
 
 
-def stack_transactions(remote_phids, local_phids):
+def stack_transactions(
+    remote_phids: List[str],
+    local_phids: List[str],
+    abandoned_revisions: Container[str],
+    no_abandon: bool = False,
+) -> Dict[str, List[Dict]]:
     """Prepare transactions to set the stack as provided in local_phids.
 
-    Returns (OrderedDict) transactions for PHID as defined in
+    Returns a dict of transactions for PHID as defined in
     `differential.revision.edit`.
     """
     remote_list = to_llist(remote_phids)
@@ -79,7 +94,10 @@ def stack_transactions(remote_phids, local_phids):
         transactions[revision] = []
 
     # Remote child no longer in stack
-    for revision in [r for r in remote_phids if r not in local_phids]:
+    remote_revisions_missing_from_local = [
+        r for r in remote_phids if r not in local_phids
+    ]
+    for revision in remote_revisions_missing_from_local:
         child = remote_list[revision]
         if child is not None:
             transactions[revision].append(("children.remove", [child]))
@@ -88,7 +106,10 @@ def stack_transactions(remote_phids, local_phids):
         walk_llist(remote_list, allow_multiple_heads=True)
 
     # New revisions
-    for revision in [r for r in local_phids if r not in remote_phids]:
+    local_revisions_missing_from_remote = [
+        r for r in local_phids if r not in remote_phids
+    ]
+    for revision in local_revisions_missing_from_remote:
         child = local_list[revision]
         if child is None:
             if revision in remote_list and remote_list[revision] is not None:
@@ -102,7 +123,8 @@ def stack_transactions(remote_phids, local_phids):
         walk_llist(remote_list, allow_multiple_heads=True)
 
     # Modify child
-    for revision in [r for r in remote_phids if r in local_phids]:
+    remote_revisions_found_in_local = [r for r in remote_phids if r in local_phids]
+    for revision in remote_revisions_found_in_local:
         if local_list[revision] != remote_list[revision]:
             child = local_list[revision]
             if child is None:
@@ -118,8 +140,16 @@ def stack_transactions(remote_phids, local_phids):
             walk_llist(remote_list, allow_multiple_heads=True)
 
     # Abandon
-    for revision in [r for r in remote_phids if r not in local_phids]:
-        transactions[revision].append(("abandon", True))
+    for revision in remote_revisions_missing_from_local:
+        if no_abandon:
+            # `--no-abandon` should cause revisions to never be abandoned.
+            logger.debug(
+                f"Skipping abandon transaction for {revision} due to `--no-abandon`."
+            )
+        elif revision not in abandoned_revisions:
+            # Avoid abandoning a revision if it is already in `abandoned` state.
+            transactions[revision].append(("abandon", True))
+
         del remote_list[revision]
         walk_llist(remote_list, allow_multiple_heads=True)
 
@@ -130,16 +160,54 @@ def stack_transactions(remote_phids, local_phids):
         if not transaction_list:
             continue
         conduit_transactions.setdefault(revision, [])
-        for transaction in transaction_list:
-            k, v = transaction
-            if k == "children.set" and v:
-                v = v
-            conduit_transactions[revision].append({"type": k, "value": v})
+
+        for trans_type, trans_value in transaction_list:
+            conduit_transactions[revision].append(
+                {"type": trans_type, "value": trans_value}
+            )
 
     return conduit_transactions
 
 
-def reorganise(repo, args):
+def convert_stackgraph_to_linear(
+    stack_graph: Dict[str, List[str]],
+    phid_to_id: Dict[str, int],
+) -> Dict[str, Optional[str]]:
+    """Converts the `stackGraph` data from Phabricator to a linear format.
+
+    Ensures each revision has only a single successor revision.
+    """
+    linear_stackgraph = {}
+
+    for successor_phid, predecessor_phid_list in stack_graph.items():
+        for predecessor_phid in predecessor_phid_list:
+            # If we find a predecessor node in our linear stackgraph then
+            # that node has multiple children.
+            if predecessor_phid in linear_stackgraph:
+                # Get the ID of the revision with multiple children.
+                rev_id = phid_to_id[predecessor_phid]
+
+                raise Error(f"Revision D{rev_id} has multiple children.")
+
+            linear_stackgraph[predecessor_phid] = successor_phid
+
+    # Set any heads as having no successor. Heads are in the values of
+    # `linear_stackgraph` but not yet added as keys.
+    heads = (
+        successor
+        # Use `list` here to avoid inspecting the list while we iterate over it,
+        # causing a `RuntimeError: dictionary changed size during iteration`.
+        for successor in list(linear_stackgraph.values())
+        if successor not in linear_stackgraph
+    )
+    for head in heads:
+        linear_stackgraph[head] = None
+
+    return linear_stackgraph
+
+
+def reorganise(repo: Repository, args: argparse.Namespace):
+    """Reorganise the stack on Phabricator to match the stack in the local VCS."""
     telemetry().submission.preparation_time.start()
 
     with wait_message("Checking connection to Phabricator."):
@@ -158,12 +226,12 @@ def reorganise(repo, args):
         augment_commits_from_body(commits)
 
     localstack_ids = [c["rev-id"] for c in commits]
-    if None in localstack_ids:
+    if not all(localstack_ids):
         names = [c["name"] for c in commits if c["rev-id"] is None]
         plural = len(names) > 1
         raise Error(
             "Found new commit{plural} in the local stack: {names}.\n"
-            "Please submit {them} separately and call the `reorg` again.".format(
+            "Please submit {them} separately and call `moz-phab reorg` again.".format(
                 plural="s" if plural else "",
                 them="them" if plural else "it",
                 names=", ".join(names),
@@ -177,18 +245,32 @@ def reorganise(repo, args):
         )
     )
 
-    # Get PhabricatorStack
-    # Errors will be raised later in the `walk_llist` method
-    with wait_message("Detecting the remote stack..."):
-        try:
-            phabstack = conduit.get_stack(localstack_ids)
-        except Error:
-            logger.error("Remote stack is not linear.")
-            raise
-
     # Preload the phabricator stack
     with wait_message("Preloading Phabricator stack revisions..."):
-        conduit.get_revisions(phids=list(phabstack.keys()))
+        revisions = conduit.get_revisions(ids=localstack_ids)
+
+    if not revisions:
+        raise Error("Could not find revisions on Phabricator.")
+
+    # Merge all the existing stackgraphs into one. Any repeated keys
+    # will have the same values.
+    stack_graph = {
+        predecessor: successors
+        for revision in revisions
+        for predecessor, successors in revision["fields"]["stackGraph"].items()
+    }
+
+    # Fetch data about any revisions that are missing from local stack.
+    revisions = conduit.get_revisions(phids=list(stack_graph.keys()))
+
+    phid_to_id = {revision["phid"]: revision["id"] for revision in revisions}
+
+    try:
+        # Validate the `stackGraph` field from our remote revisions.
+        phabstack = convert_stackgraph_to_linear(stack_graph, phid_to_id)
+    except Error:
+        logger.error("Remote stack is not linear.")
+        raise
 
     if phabstack:
         try:
@@ -205,8 +287,18 @@ def reorganise(repo, args):
         phabstack_phids = []
 
     localstack_phids = conduit.ids_to_phids(localstack_ids)
+    abandoned_revisions = {
+        revision["phid"]
+        for revision in revisions
+        if revision["fields"]["status"]["value"] == "abandoned"
+    }
     try:
-        transactions = stack_transactions(phabstack_phids, localstack_phids)
+        transactions = stack_transactions(
+            phabstack_phids,
+            localstack_phids,
+            abandoned_revisions,
+            no_abandon=args.no_abandon,
+        )
     except Error:
         logger.error("Unable to prepare stack transactions.")
         raise
@@ -217,7 +309,7 @@ def reorganise(repo, args):
     logger.warning("Stack will be reorganised:")
     for phid, rev_transactions in transactions.items():
         node_id = conduit.phid_to_id(phid)
-        if "abandon" in [t["type"] for t in rev_transactions]:
+        if any(transaction["type"] == "abandon" for transaction in rev_transactions):
             logger.info(" * {} will be abandoned".format(node_id))
         else:
             for t in rev_transactions:
@@ -249,8 +341,7 @@ def reorganise(repo, args):
 
     with wait_message("Applying transactions..."):
         for phid, rev_transactions in transactions.items():
-            conduit.edit_revision(
-                preserve_status=True,
+            conduit.apply_transactions_to_revision(
                 rev_id=phid,
                 transactions=rev_transactions,
             )
@@ -260,36 +351,42 @@ def reorganise(repo, args):
 
 
 def add_parser(parser):
-    reorg_parser = parser.add_parser("reorg", help="Reorganise commits in Phabricator")
+    reorg_parser = parser.add_parser("reorg", help="Reorganise commits in Phabricator.")
     reorg_parser.add_argument(
         "--yes",
         "-y",
         action="store_true",
-        help="Reorganise without confirmation (default: False)",
+        help="Reorganise without confirmation (default: False).",
     )
     reorg_parser.add_argument(
         "--safe-mode",
         dest="safe_mode",
         action="store_true",
-        help="Run VCS with only necessary extensions",
+        help="Run VCS with only necessary extensions.",
     )
     reorg_parser.add_argument(
         "--upstream",
         "--remote",
         "-u",
         action="append",
-        help='Set upstream branch to detect the starting commit (default: "")',
+        help='Set upstream branch to detect the starting commit (default: "").',
     )
     reorg_parser.add_argument(
         "start_rev",
         nargs="?",
         default="(auto)",
-        help="Start revision of range to reorganise (default: detected)",
+        help="Start revision of range to reorganise (default: detected).",
     )
     reorg_parser.add_argument(
         "end_rev",
         nargs="?",
         default=".",
-        help="End revision of range to reorganise (default: current commit)",
+        help="End revision of range to reorganise (default: current commit).",
+    )
+    reorg_parser.add_argument(
+        "--no-abandon",
+        action="store_true",
+        dest="no_abandon",
+        help="Do not abandon revisions during reorg.",
     )
     reorg_parser.set_defaults(func=reorganise, needs_repo=True)
