@@ -2,7 +2,9 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
-import re
+import argparse
+
+from typing import List
 
 from mozphab import environment
 
@@ -14,18 +16,19 @@ from mozphab.helpers import (
     BLOCKING_REVIEWERS_RE,
     parse_arc_diff_rev,
     prompt,
+    move_drev_to_original,
     revision_title_from_commit,
     strip_differential_revision,
     update_commit_title_previews,
 )
 from mozphab.logger import logger
+from mozphab.repository import Repository
 from mozphab.spinner import wait_message
 from mozphab.telemetry import telemetry
 
 # The DEFAULT_UPDATE_MESSAGE is only required when using arc to submit.
 DEFAULT_UPDATE_MESSAGE = "Revision updated."
 
-REVISION_URL_RE = re.compile(r"^\s*Revision URI: (http.+)$", flags=re.MULTILINE)
 PHABRICATOR_URLS = {
     "https://phabricator.services.mozilla.com/": "Phabricator",
     "https://phabricator-dev.allizom.org/": "Phabricator-Dev",
@@ -49,7 +52,7 @@ Bug #: {bug_id}
 """.strip()
 
 
-def morph_blocking_reviewers(commits):
+def morph_blocking_reviewers(commits: List[dict]):
     """Automatically fix common typo by replacing r!user with r=user!"""
 
     def morph_reviewer(matchobj):
@@ -75,14 +78,8 @@ def morph_blocking_reviewers(commits):
         commit["title"] = BLOCKING_REVIEWERS_RE.sub(morph_reviewer, commit["title"])
 
 
-def extract_revision_url(line):
-    m = REVISION_URL_RE.search(line)
-    if m:
-        return m.group(1)
-
-
-def arc_message(template_vars):
-    """Build arc commit desc message from the template"""
+def arc_message(template_vars: dict) -> str:
+    """Build arc commit desc message from the template."""
 
     # Map `None` to an empty string.
     for name in list(template_vars.keys()):
@@ -98,7 +95,7 @@ def arc_message(template_vars):
     return message
 
 
-def amend_revision_url(body, new_url):
+def amend_revision_url(body: str, new_url: str) -> str:
     """Append or replace the Differential Revision URL in a commit body."""
     body = strip_differential_revision(body)
     if body:
@@ -108,10 +105,10 @@ def amend_revision_url(body, new_url):
 
 
 def show_commit_stack(
-    commits,
-    validate=True,
-    show_rev_urls=False,
-    show_updated_only=False,
+    commits: List[dict],
+    validate: bool = True,
+    show_rev_urls: bool = False,
+    show_updated_only: bool = False,
 ):
     """Log the commit stack in a human readable form."""
 
@@ -130,7 +127,9 @@ def show_commit_stack(
 
             # preload diffs
             with wait_message("Loading diffs..."):
-                diffs = conduit.get_diffs([r["fields"]["diffPHID"] for r in revisions])
+                diffs = conduit.get_diffs(
+                    phids=[r["fields"]["diffPHID"] for r in revisions]
+                )
 
     for commit in reversed(commits):
         if show_updated_only and not commit["submit"]:
@@ -264,11 +263,12 @@ def show_commit_stack(
             logger.warning("-> %s/D%s", conduit.repo.phab_url, commit["rev-id"])
 
 
-def make_blocking(reviewers):
+def make_blocking(reviewers: List[str]) -> List[str]:
+    """Convert a list of reviewer strings into a list of blocking reviewer strings."""
     return ["%s!" % r.rstrip("!") for r in reviewers]
 
 
-def remove_duplicates(reviewers):
+def remove_duplicates(reviewers: List[str]) -> List[str]:
     """Remove all duplicate items from the list.
 
     Args:
@@ -293,12 +293,12 @@ def remove_duplicates(reviewers):
     return unique
 
 
-def update_commits_from_args(commits, args):
+def update_commits_from_args(commits: List[dict], args: argparse.Namespace):
     """Modify commit description based on args and configuration.
 
     Args:
         commits: list of dicts representing the commits in commit stack
-        args: argparse.ArgumentParser object. In this function we're using
+        args: argparse.Namespace object. In this function we're using
             following attributes:
             * reviewer - list of strings representing reviewers
             * blocker - list of string representing blocking reviewers
@@ -386,10 +386,30 @@ def update_commits_from_args(commits, args):
                 granted=make_blocking(commit["reviewers"]["granted"]),
             )
 
+    if args.command == "uplift":
+        update_commits_for_uplift(commits)
+
     update_commit_title_previews(commits)
 
 
-def update_revision_description(transactions, commit, revision):
+def update_commits_for_uplift(commits: List[dict]):
+    """Prepares a set of commits for uplifting."""
+    for commit in commits:
+        # When uplifting, ensure the `Differential Revision` line is properly
+        # moved, and that `rev-id` is updated to not point at the original revision.
+        commit["body"], commit["rev-id"] = move_drev_to_original(
+            commit["body"],
+            commit["rev-id"],
+        )
+
+        # Clear all reviewers from the revision.
+        commit["reviewers"] = {
+            "granted": [],
+            "request": [],
+        }
+
+
+def update_revision_description(transactions: List[dict], commit: dict, revision: dict):
     # Appends differential.revision.edit transaction(s) to `transactions` if
     # updating the commit title and/or summary is required.
 
@@ -404,15 +424,56 @@ def update_revision_description(transactions, commit, revision):
         transactions.append(dict(type="summary", value=local_body))
 
 
-def update_revision_bug_id(transactions, commit, revision):
+def update_revision_bug_id(transactions: List[dict], commit: dict, revision: dict):
     # Appends differential.revision.edit transaction(s) to `transactions` if
     # updating the commit bug-id is required.
     if commit["bug-id"] and commit["bug-id"] != revision["fields"]["bugzilla.bug-id"]:
         transactions.append(dict(type="bugzilla.bug-id", value=commit["bug-id"]))
 
 
-def submit(repo, args):
-    telemetry.metrics.mozphab.submission.preparation_time.start()
+def local_uplift_if_possible(
+    args: argparse.Namespace, repo: Repository, commits: List[dict]
+) -> bool:
+    """If possible, rebase local repository commits onto the target uplift train.
+
+    Uplifts will be performed if the `--no-rebase` argument is not
+    passed, if the local repository has a corresponding unified head
+    for the uplift train, and if the revset being submitted is not
+    already descendant from the target unified head.
+
+    Returns a `bool` indicating if the `commits` should avoid making local
+    repository changes to reflect new Phabricator revisions. If `True`,
+    local commits may be amended to reflect their state on Phabricator,
+    for example to add the `Differential Revision` line. If `False`, the
+    local repository commits should not be updated by `moz-phab`.
+    """
+    if args.no_rebase:
+        # If args tell us not to do a rebase, do not make any local changes and
+        # return without rebasing. This is the same as submitting an uplift where
+        # the original patch is sent to Phabricator without any modifications.
+        # In this case we want to avoid local amendments to commits.
+        return True
+
+    # Try and find a local repo identifier (hg bookmark, git remote branch) to rebase
+    # our revset onto.
+    unified_head = repo.map_callsign_to_unified_head(args.train)
+
+    if not unified_head:
+        # If we didn't find a unified head, we intend to submit an uplift without
+        # modifying the local repo state via a rebase.
+        return True
+
+    if not repo.is_descendant(unified_head):
+        # If we found a head to rebase onto and the commit isn't already a descendant
+        # of our target identifier, uplift the commits onto the unified head.
+        with wait_message(f"Rebasing commits onto {unified_head}"):
+            commits = repo.uplift_commits(unified_head, commits)
+
+    return False
+
+
+def _submit(repo: Repository, args: argparse.Namespace):
+    telemetry().submission.preparation_time.start()
     with wait_message("Checking connection to Phabricator."):
         # Check if raw Conduit API can be used
         if not conduit.check():
@@ -429,6 +490,12 @@ def submit(repo, args):
     if not commits:
         raise Error("Failed to find any commits to submit")
 
+    if args.command == "uplift":
+        # Perform uplift logic during submission.
+        avoid_local_changes = local_uplift_if_possible(args, repo, commits)
+    else:
+        avoid_local_changes = False
+
     with wait_message("Loading commits.."):
         # Pre-process to load metadata.
         morph_blocking_reviewers(commits)
@@ -479,11 +546,11 @@ def submit(repo, args):
     if args.message and any([c for c in commits if not c["rev-id"]]):
         logger.warning(
             "Warning: --message works with updates only, and will not\n"
-            "be result in a comment on new revisions."
+            "result in a comment on new revisions."
         )
 
-    telemetry.metrics.mozphab.submission.preparation_time.stop()
-    telemetry.metrics.mozphab.submission.commits_count.add(len(commits))
+    telemetry().submission.preparation_time.stop()
+    telemetry().submission.commits_count.add(len(commits))
 
     # Confirmation prompt.
     if args.yes:
@@ -504,7 +571,7 @@ def submit(repo, args):
             config.write()
 
     # Process.
-    telemetry.metrics.mozphab.submission.process_time.start()
+    telemetry().submission.process_time.start()
     previous_commit = None
     # Collect all existing revisions to get reviewers info.
     rev_ids_to_update = [int(c["rev-id"]) for c in commits if c.get("rev-id")]
@@ -529,10 +596,10 @@ def submit(repo, args):
         revision_to_update = (
             revisions_to_update[commit["rev-id"]] if is_update else None
         )
-        existing_reviewers = (
-            revision_to_update["attachments"]["reviewers"]["reviewers"]
+        has_existing_reviewers = (
+            bool(revision_to_update["attachments"]["reviewers"]["reviewers"])
             if revision_to_update
-            else None
+            else False
         )
 
         # Let the user know something's happening.
@@ -571,7 +638,7 @@ def submit(repo, args):
             diff = repo.get_diff(commit)
 
         if diff:
-            telemetry.metrics.mozphab.submission.files_count.add(len(diff.changes))
+            telemetry().submission.files_count.add(len(diff.changes))
             with wait_message("Uploading binary file(s)..."):
                 diff.upload_files()
 
@@ -582,7 +649,7 @@ def submit(repo, args):
             with wait_message("Updating revision..."):
                 rev = conduit.update_revision(
                     commit,
-                    existing_reviewers,
+                    has_existing_reviewers,
                     diff_phid=diff.phid,
                     comment=args.message,
                     check_in_needed=check_in_needed,
@@ -608,8 +675,10 @@ def submit(repo, args):
             commit["title"] = commit["title-preview"]
             commit["body"] = body
             commit["rev-id"] = parse_arc_diff_rev(commit["body"])
-            with wait_message("Updating commit.."):
-                repo.amend_commit(commit, commits)
+
+            if not avoid_local_changes:
+                with wait_message("Updating commit.."):
+                    repo.amend_commit(commit, commits)
 
         # Diff property has to be set after potential SHA1 change.
         if diff:
@@ -630,90 +699,110 @@ def submit(repo, args):
     show_commit_stack(
         commits, validate=False, show_rev_urls=True, show_updated_only=True
     )
-    telemetry.metrics.mozphab.submission.process_time.stop()
+    telemetry().submission.process_time.stop()
+
+
+def submit(repo: Repository, args: argparse.Namespace):
+    try:
+        _submit(repo, args)
+    except Exception as e:
+        # Check if the `fallback` flag was set and print a warning message.
+        if hasattr(args, "fallback") and args.fallback:
+            logger.warning(
+                "You didn't specify a valid command, so we ran `submit` for you, "
+                "and it failed."
+            )
+
+        raise e
 
 
 def add_parser(parser):
     submit_parser = parser.add_parser(
         "submit",
         aliases=["upload"],
-        help="Submit commit(s) to Phabricator",
+        help="Submit commit(s) to Phabricator.",
         description=(
             "MozPhab will change the working directory and amend the commits during "
             "the submission process."
         ),
     )
-    submit_parser.add_argument(
-        "--path", "-p", help="Set path to repository (default: detected)"
+    add_submit_arguments(submit_parser)
+    submit_parser.set_defaults(func=submit, needs_repo=True)
+
+
+def add_submit_arguments(parser):
+    """Add `moz-phab submit` command line arguments to a parser."""
+    parser.add_argument(
+        "--path", "-p", help="Set path to repository (default: detected)."
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--yes",
         "-y",
         action="store_true",
-        help="Submit without confirmation (default: %s)" % config.auto_submit,
+        help="Submit without confirmation (default: %s)." % config.auto_submit,
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--interactive",
         "-i",
         action="store_true",
-        help="Submit with confirmation (default: %s)" % (not config.auto_submit),
+        help="Submit with confirmation (default: %s)." % (not config.auto_submit),
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--message",
         "-m",
-        help="Provide a custom update message (default: none)",
+        help="Provide a custom update message (default: none).",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--force",
         "-f",
         action="store_true",
-        help="Override sanity checks and force submission; a tool of last resort",
+        help="Override sanity checks and force submission; a tool of last resort.",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--force-delete",
         action="store_true",
         help="Mercurial only: Ignore error caused by a DAG branch point without "
-        "evolve installed",
+        "evolve installed.",
     )
-    submit_parser.add_argument(
-        "--bug", "-b", help="Set Bug ID for all commits (default: from commit)"
+    parser.add_argument(
+        "--bug", "-b", help="Set Bug ID for all commits (default: from commit)."
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--no-bug",
         action="store_true",
-        help="Continue if a bug number is not provided",
+        help="Continue if a bug number is not provided.",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--reviewer",
         "--reviewers",
         "-r",
         action="append",
-        help="Set review(s) for all commits (default: from commit)",
+        help="Set review(s) for all commits (default: from commit).",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--blocker",
         "--blockers",
         "-R",
         action="append",
-        help="Set blocking review(s) for all commits (default: from commit)",
+        help="Set blocking review(s) for all commits (default: from commit).",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--nolint",
         "--no-lint",
         action="store_true",
-        help="Do not run lint (default: lint changed files if configured)",
+        help="Do not run lint (default: lint changed files if configured).",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--check-in-needed",
         action="store_true",
-        help="Add a `check-in-needed tag to the top most revision",
+        help="Add a `check-in-needed tag to the top most revision.",
     )
-    wip_group = submit_parser.add_mutually_exclusive_group()
+    wip_group = parser.add_mutually_exclusive_group()
     wip_group.add_argument(
         "--wip",
         "--plan-changes",
         action="store_true",
-        help="Create or update a revision without requesting a code review",
+        help="Create or update a revision without requesting a code review.",
     )
     wip_group.add_argument(
         "--no-wip",
@@ -722,7 +811,7 @@ def add_parser(parser):
         "with descriptions that start with WIP: will continue to be "
         "flagged as work-in-progress.",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--less-context",
         "--lesscontext",
         action="store_true",
@@ -735,46 +824,44 @@ def add_parser(parser):
             "revision will be created that has only a few lines of context."
         ),
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--no-stack",
         action="store_true",
-        help="Submit multiple commits, but do not mark them as dependent",
+        help="Submit multiple commits, but do not mark them as dependent.",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--upstream",
         "--remote",
         "-u",
         action="append",
-        help='Set upstream branch to detect the starting commit (default: "")',
+        help='Set upstream branch to detect the starting commit (default: "").',
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--force-vcs",
         action="store_true",
-        help="EXPERIMENTAL: Override VCS compatibility check",
+        help="EXPERIMENTAL: Override VCS compatibility check.",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--safe-mode",
         dest="safe_mode",
         action="store_true",
-        help="Run VCS with only necessary extensions",
+        help="Run VCS with only necessary extensions.",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "--single",
         "-s",
         action="store_true",
-        help="Submit a single commit",
+        help="Submit a single commit.",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "start_rev",
         nargs="?",
         default=environment.DEFAULT_START_REV,
-        help="Start revision of range to submit (default: detected)",
+        help="Start revision of range to submit (default: detected).",
     )
-    submit_parser.add_argument(
+    parser.add_argument(
         "end_rev",
         nargs="?",
         default=environment.DEFAULT_END_REV,
-        help="End revision of range to submit (default: current commit)",
+        help="End revision of range to submit (default: current commit).",
     )
-
-    submit_parser.set_defaults(func=submit, needs_repo=True)
