@@ -10,6 +10,13 @@ import os
 import urllib.parse as url_parse
 import urllib.request as url_request
 
+from typing import (
+    Any,
+    Dict,
+    List,
+    Optional,
+)
+
 from mozphab import environment
 
 from .environment import USER_AGENT
@@ -32,7 +39,7 @@ from .simplecache import cache
 CHECK_IN_NEEDED = "check-in_needed"
 
 
-def normalise_reviewer(reviewer, strip_group=True):
+def normalise_reviewer(reviewer: str, strip_group: bool = True) -> str:
     """This provide a canonical form of the reviewer for comparison."""
     reviewer = reviewer.rstrip("!").lower()
     if strip_group:
@@ -43,7 +50,7 @@ def normalise_reviewer(reviewer, strip_group=True):
 class ConduitAPIError(Error):
     """Raised when the Phabricator Conduit API returns an error response."""
 
-    def __init__(self, msg=None):
+    def __init__(self, msg: Optional[str] = None):
         super().__init__(f"Phabricator Error: {msg if msg else 'Unknown Error'}")
 
 
@@ -55,10 +62,10 @@ class ConduitAPI:
         self.repo = repo
 
     @property
-    def repo_phid(self):
+    def repo_phid(self) -> str:
         return self.repo.phid
 
-    def load_api_token(self):
+    def load_api_token(self) -> str:
         """Return an API Token for the given repository.
 
         Returns:
@@ -66,7 +73,7 @@ class ConduitAPI:
         """
 
         if "api_token" in cache:
-            return cache.get("api_token")
+            return str(cache.get("api_token"))
 
         token = read_json_field(
             [get_arcrc_path()], ["hosts", self.repo.api_url, "token"]
@@ -76,7 +83,7 @@ class ConduitAPI:
         cache.set("api_token", token)
         return token
 
-    def save_api_token(self, token):
+    def save_api_token(self, token: str):
         filename = get_arcrc_path()
         created = False
         try:
@@ -96,7 +103,9 @@ class ConduitAPI:
         if created:
             os.chmod(filename, 0o600)
 
-    def call(self, api_method, api_call_args, *, api_token=None):
+    def call(
+        self, api_method: str, api_call_args: dict, *, api_token: Optional[str] = None
+    ) -> dict:
         """Call Conduit API and return the JSON API call result.
 
         Args:
@@ -123,7 +132,9 @@ class ConduitAPI:
             raise ConduitAPIError(res.get("error_info", "Error %s" % res["error_code"]))
         return res["result"]
 
-    def _build_request(self, *, method, args, token):
+    def _build_request(
+        self, *, method: str, args: dict, token: Optional[str]
+    ) -> Dict[str, Any]:
         """Return dict with Request args for calling the specified conduit method."""
         return dict(
             url=url_parse.urljoin(self.repo.api_url, method),
@@ -144,7 +155,7 @@ class ConduitAPI:
             ).encode(),
         )
 
-    def ping(self):
+    def ping(self) -> bool:
         """Sends a ping to the Phabricator server using `conduit.ping` API.
 
         Returns: `True` if no error, otherwise - `False`
@@ -159,7 +170,7 @@ class ConduitAPI:
             return False
         return True
 
-    def check(self):
+    def check(self) -> bool:
         """Check if raw Conduit API can be used."""
         # Check if the cache file exists
         path = os.path.join(self.repo.dot_path, ".moz-phab_conduit-configured")
@@ -174,19 +185,19 @@ class ConduitAPI:
 
         return False
 
-    def get_projects(self, slugs):
+    def get_projects(self, slugs: List[str]) -> List[dict]:
         """Search for tags by hashtags."""
         response = self.call("project.search", dict(constraints=dict(slugs=slugs)))
         return response.get("data")
 
-    def get_project_phid(self, slug):
+    def get_project_phid(self, slug: str) -> Optional[str]:
         projects = self.get_projects([slug])
         if not projects:
             return None
 
         return projects[0]["phid"]
 
-    def ids_to_phids(self, rev_ids):
+    def ids_to_phids(self, rev_ids: List[int]) -> List[str]:
         """Convert revision ids to PHIDs.
 
         Parameters:
@@ -197,7 +208,7 @@ class ConduitAPI:
         """
         return [r["phid"] for r in self.get_revisions(ids=rev_ids)]
 
-    def id_to_phid(self, rev_id):
+    def id_to_phid(self, rev_id: int) -> str:
         """Convert revision id to PHID."""
         phids = self.ids_to_phids([rev_id])
         if phids:
@@ -205,7 +216,7 @@ class ConduitAPI:
 
         raise NotFoundError("revision {} not found".format(rev_id))
 
-    def phids_to_ids(self, phids):
+    def phids_to_ids(self, phids: List[str]) -> List[str]:
         """Convert revision PHIDs to ids.
 
         Parameteres:
@@ -216,7 +227,7 @@ class ConduitAPI:
         """
         return ["D{}".format(r["id"]) for r in self.get_revisions(phids=phids)]
 
-    def phid_to_id(self, phid):
+    def phid_to_id(self, phid: str) -> str:
         """Convert revision PHID to id."""
         ids = self.phids_to_ids([phid])
         if ids:
@@ -224,7 +235,9 @@ class ConduitAPI:
 
         raise NotFoundError("revision {} not found".format(phid))
 
-    def get_revisions(self, ids=None, phids=None):
+    def get_revisions(
+        self, ids: Optional[List[int]] = None, phids: Optional[List[int]] = None
+    ) -> List[dict]:
         """Get revisions info from Phabricator.
 
         Args:
@@ -293,16 +306,27 @@ class ConduitAPI:
         else:
             return [revisions[phid] for phid in phids]
 
-    def get_diffs(self, phids):
+    def get_diffs(
+        self, ids: Optional[List[int]] = None, phids: Optional[List[str]] = None
+    ) -> Dict[str, Dict]:
         """Get diffs from Phabricator.
 
         Args:
+            ids - a list of diff IDs to pull
             phids - a list of diff PHIDs to pull
 
         Returns a dict of diffs identified by their PHID
         """
+        if (ids and phids) or (ids is None and phids is None):
+            raise ValueError("Internal Error: Invalid args to get_diffs")
+
+        if ids:
+            constraints = {"ids": list(set(ids))}
+        else:
+            constraints = {"phids": list(set(phids))}
+
         api_call_args = {
-            "constraints": {"phids": phids},
+            "constraints": constraints,
             "attachments": {"commits": True},
         }
         response = self.call("differential.diff.search", api_call_args)
@@ -314,17 +338,23 @@ class ConduitAPI:
 
         return diff_dict
 
-    def get_successor_phids(self, phid, include_abandoned=False):
+    def get_successor_phids(
+        self, phid: str, include_abandoned: bool = False
+    ) -> List[str]:
         return self.get_related_phids(
             phid, relation="child", include_abandoned=include_abandoned
         )
 
-    def get_ancestor_phids(self, phid, include_abandoned=False):
+    def get_ancestor_phids(
+        self, phid: str, include_abandoned: bool = False
+    ) -> List[str]:
         return self.get_related_phids(
             phid, relation="parent", include_abandoned=include_abandoned
         )
 
-    def get_related_phids(self, base_phid, relation="parent", include_abandoned=False):
+    def get_related_phids(
+        self, base_phid: str, relation: str = "parent", include_abandoned: bool = False
+    ) -> List[str]:
         """Returns the list of PHIDs with direct dependency"""
         result = []
 
@@ -350,52 +380,7 @@ class ConduitAPI:
             if r["fields"]["status"]["value"] != "abandoned"
         ]
 
-    def get_stack(self, rev_ids):
-        """Returns a dict of PHIDs."""
-        phids = set()
-        if not rev_ids:
-            return {}
-        revisions = self.get_revisions(ids=rev_ids)
-        new_phids = set([rev["phid"] for rev in revisions])
-        stack = {}
-
-        while new_phids:
-            phids.update(new_phids)
-
-            edges = self.call(
-                "edge.search",
-                dict(
-                    sourcePHIDs=list(new_phids),
-                    types=["revision.parent", "revision.child"],
-                    limit=10000,
-                ),
-            )["data"]
-
-            new_phids = set()
-            for edge in edges:
-                new_phids.add(edge["sourcePHID"])
-                new_phids.add(edge["destinationPHID"])
-
-                if edge["edgeType"] == "revision.child":
-                    if edge["sourcePHID"] in stack:
-                        source_id = next(
-                            r["id"]
-                            for r in revisions
-                            if r["phid"] == edge["sourcePHID"]
-                        )
-                        raise Error("Revision D%s has multiple children." % source_id)
-
-                    stack[edge["sourcePHID"]] = edge["destinationPHID"]
-
-            new_phids = new_phids - phids
-
-        for child in list(stack.values()):
-            # set the last child (not a parent)
-            stack.setdefault(child)
-
-        return stack
-
-    def get_users(self, usernames):
+    def get_users(self, usernames: List[str]) -> List[dict]:
         """Get users using the user.query API.
 
         Caches the result in the process.
@@ -427,7 +412,7 @@ class ConduitAPI:
 
         return users
 
-    def get_groups(self, slugs):
+    def get_groups(self, slugs: List[str]) -> List[dict]:
         to_collect = []
         groups = []
         for slug in slugs:
@@ -464,11 +449,11 @@ class ConduitAPI:
 
     def create_revision(
         self,
-        commit,
-        summary,
-        diff_phid,
-        check_in_needed=False,
-    ):
+        commit: dict,
+        summary: str,
+        diff_phid: str,
+        check_in_needed: bool = False,
+    ) -> dict:
         """Create a new revision in Phabricator."""
         transactions = [
             dict(type="title", value=revision_title_from_commit(commit)),
@@ -488,12 +473,12 @@ class ConduitAPI:
 
     def update_revision(
         self,
-        commit,
-        existing_reviewers,
-        diff_phid=None,
-        comment=None,
-        check_in_needed=False,
-    ):
+        commit: dict,
+        has_existing_reviewers: bool,
+        diff_phid: Optional[str] = None,
+        comment: Optional[str] = None,
+        check_in_needed: bool = False,
+    ) -> dict:
         """Update an existing revision in Phabricator."""
         # Update the title and summary
         transactions = [
@@ -507,7 +492,7 @@ class ConduitAPI:
 
         # Add reviewers only if revision lacks them
         if commit["has-reviewers"] and not commit["wip"]:
-            if not existing_reviewers:
+            if not has_existing_reviewers:
                 self.update_revision_reviewers(transactions, commit)
 
         # Update bug id if different
@@ -528,15 +513,15 @@ class ConduitAPI:
 
     def edit_revision(
         self,
-        transactions=None,
-        diff_phid=None,
-        rev_id=None,
-        wip=False,
-        check_in_needed=False,
-    ):
+        transactions: Optional[List[dict]] = None,
+        diff_phid: Optional[str] = None,
+        rev_id: Optional[str] = None,
+        wip: bool = False,
+        check_in_needed: bool = False,
+    ) -> dict:
         """Edit (create or update) a revision."""
-        trans = list(transactions or [])
-        set_wip_later = False
+        trans = transactions or []
+        post_trans = []
 
         # diff_phid is not present for changes in revision settings (like WIP)
         if diff_phid:
@@ -555,14 +540,16 @@ class ConduitAPI:
         # Phabricator will throw an error if we attempt to set a status to the same
         # as the current status.
         if wip:
-            # Phabriactor will automatically set the revision to needs-review
-            # after the call to differential.revision.edit.  Flag that we'll need to
-            # make a subsequent API call to set the status to changes-planned to match
-            # our WIP state.
+            # Phabricator will automatically set the revision to needs-review
+            # after the call to differential.revision.edit. If we are a creating a new
+            # revision, we can add the `plan-changes` transaction to our create call.
+            # Existing revisions must make a subsequent API call to set the status to
+            # `changes-planned` to match our WIP state.
+            plan_changes_transaction = dict(type="plan-changes", value=True)
             if existing_status == "changes-planned":
-                set_wip_later = True
+                post_trans.append(plan_changes_transaction)
             else:
-                trans.append(dict(type="plan-changes", value=True))
+                trans.append(plan_changes_transaction)
 
         # Let the Phabricator sort out the correct status for new revisions to avoid
         # emails being sent before phabbugs has processed the revision.
@@ -591,35 +578,63 @@ class ConduitAPI:
         if not revision:
             raise ConduitAPIError("Can't edit the revision.")
 
-        # Set changes-planned if required.
-        # Note set_wip_later can only be true if there's an existing revision.
-        if set_wip_later:
+        # Run post-edit transactions if required.
+        if post_trans:
             revision = self.call(
                 "differential.revision.edit",
                 dict(
                     objectIdentifier=rev_id,
-                    transactions=[dict(type="plan-changes", value=True)],
+                    transactions=post_trans,
                 ),
             )
 
         return revision
 
-    def get_repository(self, call_sign):
-        """Get the repository info from Phabricator."""
-        key = "repo-%s" % call_sign
-        if key in cache:
-            return cache.get(key)
+    def apply_transactions_to_revision(self, rev_id: str, transactions: List[dict]):
+        """Apply transactions to the specified revision."""
+        self.call(
+            "differential.revision.edit",
+            {"objectIdentifier": rev_id, "transactions": transactions},
+        )
 
-        api_call_args = dict(constraints=dict(callsigns=[call_sign]), limit=1)
+    def get_repository_by_callsign(self, call_sign: str) -> dict:
+        """Get repository info for a repo on Phabricator by callsign."""
+        return self.repository_search_single("callsigns", call_sign)
+
+    def get_repository_by_shortname(self, short_name: str) -> dict:
+        """Get repository info for a repo on Phabricator by shortname."""
+        return self.repository_search_single("shortNames", short_name)
+
+    def repository_search_single(self, constraint: str, value: str) -> dict:
+        """Get the information about a single repository from Phabricator.
+
+        Takes a `constraint` to be passed to `diffusion.repository.search`
+        and a `value` to be passed as the value of `constraint`. `value`
+        must be a single element and will be passed in a list to Conduit.
+        """
+        api_call_args = dict(constraints={constraint: [value]}, limit=1)
         data = self.call("diffusion.repository.search", api_call_args)
         if not data.get("data"):
-            raise NotFoundError("Repository %s not found" % call_sign)
+            raise NotFoundError("Repository %s not found" % value)
 
         repo = data["data"][0]
-        cache.set(key, repo)
         return repo
 
-    def create_diff(self, changes, base_revision):
+    def get_repositories_with_tag(self, tag: str) -> dict:
+        """Get repository information for repos associated with the given tag."""
+        api_call_args = {
+            "constraints": {
+                "projects": [tag],
+            }
+        }
+
+        data = self.call("diffusion.repository.search", api_call_args).get("data")
+        if not data:
+            raise NotFoundError(f"No repositories found with tag {tag}")
+
+        return data
+
+    def create_diff(self, changes: List[Dict[str, Any]], base_revision: str) -> dict:
         creation_method = ["moz-phab", conduit.repo.vcs]
         if conduit.repo.vcs == "git" and conduit.repo.is_cinnabar_required:
             creation_method.append("cinnabar")
@@ -639,7 +654,7 @@ class ConduitAPI:
         )
         return self.call("differential.creatediff", api_call_args)
 
-    def set_diff_property(self, diff_id, commit, message):
+    def set_diff_property(self, diff_id: str, commit: dict, message: str):
         data = {
             commit["node"]: {
                 "author": commit["author-name"],
@@ -662,7 +677,7 @@ class ConduitAPI:
         )
         self.call("differential.setdiffproperty", api_call_args)
 
-    def file_upload(self, path, data):
+    def file_upload(self, path: str, data: bytes) -> Optional[str]:
         if not data:
             return
         name = os.path.basename(path)
@@ -699,17 +714,19 @@ class ConduitAPI:
                         ),
                     )
 
-        return file_phid
+        return str(file_phid)
 
-    def whoami(self, *, api_token=None):
+    def whoami(self, *, api_token: Optional[str] = None) -> dict:
         if "whoami" in cache:
-            return cache.get("whoami")
+            return dict(cache.get("whoami"))
 
         who = self.call("user.whoami", {}, api_token=api_token)
         cache.set("whoami", who)
         return who
 
-    def update_revision_reviewers(self, transactions, commit):
+    def update_revision_reviewers(
+        self, transactions: List[Dict[str, Any]], commit: dict
+    ):
         # Appends differential.revision.edit transaction(s) to `transactions` to
         # set the reviewers.
 
@@ -742,7 +759,7 @@ class ConduitAPI:
         )
         transactions.extend([dict(type="reviewers.set", value=all_reviewing_phid)])
 
-    def check_for_invalid_reviewers(self, reviewers):
+    def check_for_invalid_reviewers(self, reviewers: dict) -> List[Dict[str, Any]]:
         """Return a list of invalid reviewer names.
 
         Args:
