@@ -9,7 +9,7 @@ class Error(Exception):
 
 
 class NotFoundError(Exception):
-    """Errors raised when node is not found."""
+    """Errors raised when data is not found."""
 
 
 class NonLinearException(Exception):
@@ -19,7 +19,7 @@ class NonLinearException(Exception):
 class CommandError(Exception):
     """Errors raised by external commands."""
 
-    status = None
+    status: int
 
     def __init__(self, msg="", status=1):
         self.status = status
