@@ -35,7 +35,12 @@ ARC_REJECT_RE_LIST = [
 
 
 ARC_DIFF_REV_RE = re.compile(
-    r"^\s*Differential Revision:\s*https?://.+/D(\d+)\s*$", flags=re.MULTILINE
+    r"^\s*Differential Revision:\s*(?P<phab_url>https?://.+)/D(?P<rev>\d+)\s*$",
+    flags=re.MULTILINE,
+)
+ORIGINAL_DIFF_REV_RE = re.compile(
+    r"^\s*Original Revision:\s*(?P<phab_url>https?://.+)/D(?P<rev>\d+)\s*$",
+    flags=re.MULTILINE,
 )
 
 # Bug and review regexs (from vct's commitparser)
@@ -64,6 +69,8 @@ DEPENDS_ON_RE = re.compile(r"^\s*Depends on\s*D(\d+)\s*$", flags=re.MULTILINE)
 
 WIP_RE = re.compile(r"^(?:WIP[: ]|WIP$)", flags=re.IGNORECASE)
 
+VALID_EMAIL_RE = re.compile(r"[^@ \t\r\n]+@[^@ \t\r\n]+\.[^@ \t\r\n]+")
+
 
 def which_path(path):
     """Check if an executable is provided. Fall back to which if not.
@@ -111,7 +118,7 @@ def parse_config(config_list, filter_func=None):
     return result
 
 
-def read_json_field(files, field_path):
+def read_json_field(files: List[str], field_path: List[str]) -> Optional[str]:
     """Parses json files in turn returning value as per field_path, or None."""
     for filename in files:
         try:
@@ -156,10 +163,10 @@ def temporary_binary_file(content):
         os.remove(f.name)
 
 
-def get_arcrc_path():
+def get_arcrc_path() -> str:
     """Return a path to the user's Arcanist configuration file."""
     if "arcrc" in cache:
-        return cache.get("arcrc")
+        return str(cache.get("arcrc"))
 
     if environment.IS_WINDOWS:
         arcrc = os.path.join(os.getenv("APPDATA", ""), ".arcrc")
@@ -175,13 +182,51 @@ def get_arcrc_path():
 
 def parse_arc_diff_rev(body):
     m = ARC_DIFF_REV_RE.search(body)
-    return m.group(1) if m else None
+    return m.group("rev") if m else None
 
 
 def strip_differential_revision(body):
     return ARC_DIFF_REV_RE.sub("", body).rstrip()
 
 
+def move_drev_to_original(
+    body: str, rev_id: Optional[int]
+) -> Tuple[str, Optional[int]]:
+    """Handle moving the `Differential Revision` line.
+
+    Moves the `Differential Revision` line to `Original Revision`, if a link
+    to the original revision does not already exist. If the `Original Revision`
+    line does exist, scrub the `Differential Revision` line.
+
+    Args:
+        body: `str` text of the commit message.
+        rev_id: `int` parsed integer representing the drev number for the revision.
+
+    Returns:
+        tuple of:
+            New commit message body text as `str`,
+            Revision id as `int`, or `None` if a new revision should be created.
+    """
+    # Previous logic did not find a revision id, so this function won't either.
+    if not rev_id:
+        return body, rev_id
+
+    differential_revision = ARC_DIFF_REV_RE.search(body)
+    original_revision = ORIGINAL_DIFF_REV_RE.search(body)
+
+    # If both match, this is an update to an uplift.
+    if differential_revision and original_revision:
+        return body, rev_id
+
+    def repl(match):
+        phab_url = match.group("phab_url")
+        rev = match.group("rev")
+        return f"\nOriginal Revision: {phab_url}/D{rev}"
+
+    # Update the commit message and set the `rev-id` to `None`.
+    return ARC_DIFF_REV_RE.sub(repl, body), None
+
+
 def parse_api_error(api_response):
     """Parse the string output from 'arc call-conduit' and return any errors found.
 
@@ -304,7 +349,7 @@ def strip_depends_on(body):
     return DEPENDS_ON_RE.sub("", body).rstrip()
 
 
-def revision_title_from_commit(commit):
+def revision_title_from_commit(commit: dict) -> str:
     """Returns a string suitable for a Revision title for the given commit."""
     title = WIP_RE.sub("", commit["title-preview"]).lstrip()
     if commit["wip"]:
@@ -516,3 +561,8 @@ def join_lineseps(lines):
     return [
         "".join(line) for line in zip_longest(lines[0::2], lines[1::2], fillvalue="")
     ]
+
+
+def is_valid_email(email: str) -> bool:
+    """Given a string, determines if it is a valid email."""
+    return VALID_EMAIL_RE.match(email) is not None
