@@ -2,6 +2,7 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
+import json
 import mimetypes
 import os
 import re
@@ -10,8 +11,12 @@ import time
 import uuid
 
 from contextlib import suppress
-from distutils.version import LooseVersion
+from packaging.version import Version
 from functools import lru_cache
+from typing import (
+    List,
+    Optional,
+)
 
 import hglib
 from mozphab import environment
@@ -26,6 +31,7 @@ from .helpers import (
     temporary_binary_file,
     temporary_file,
     which_path,
+    is_valid_email,
 )
 from .logger import logger
 from .repository import Repository
@@ -33,7 +39,7 @@ from .spinner import wait_message, clear_terminal_line
 from .subprocess_wrapper import debug_log_command
 from .telemetry import telemetry
 
-MINIMUM_MERCURIAL_VERSION = LooseVersion("4.3.3")
+MINIMUM_MERCURIAL_VERSION = Version("4.3.3")
 
 
 class Mercurial(Repository):
@@ -52,10 +58,12 @@ class Mercurial(Repository):
         self.obsstore = None
         self.unlink_obsstore = False
         self.use_evolve = False
+        self.use_topic = False
         self.has_mq = False
         self.has_shelve = False
         self.previous_bookmark = None
         self.has_temporary_bookmark = False
+        self.username = ""
 
         # Check for `hg` presence
         if not which_path(self._hg_binary):
@@ -70,7 +78,7 @@ class Mercurial(Repository):
         self._repo = None
         self._configs = []
         major, minor, micro, *_ = self.repository.version
-        self.mercurial_version = LooseVersion(f"{major}.{minor}.{micro}")
+        self.mercurial_version = Version(f"{major}.{minor}.{micro}")
         self.vcs_version = str(self.mercurial_version)
         if self.mercurial_version < MINIMUM_MERCURIAL_VERSION:
             raise Error(
@@ -211,6 +219,8 @@ class Mercurial(Repository):
         return self.hg_out(["log", "-T", "{%s}\n" % select, "-r", revset], split=split)
 
     def before_submit(self):
+        self.validate_email()
+
         # Remember the currently checked out commit.  If a bookmark is active
         # just use that, otherwise create a randomly named bookmark which will
         # be deleted in cleanup(). Mercurial will automatically move the
@@ -305,8 +315,6 @@ class Mercurial(Repository):
 
         self.revset = "%s::%s" % (commits[0]["node"], commits[-1]["node"])
 
-        super().refresh_commit_stack(commits)
-
     def set_args(self, args):
         """Sets up the right environment for hg, prior to running it.
 
@@ -336,6 +344,7 @@ class Mercurial(Repository):
         if "ui.username" not in hg_config:
             raise Error("ui.username is not configured in your hgrc")
         self._safe_config_options["ui.username"] = hg_config["ui.username"]
+        self.username = hg_config["ui.username"]
 
         # Always need rebase.
         self._config_options["extensions.rebase"] = ""
@@ -344,7 +353,7 @@ class Mercurial(Repository):
         self._extra_options["--pager"] = "never"
 
         # Perform rebases in-memory to improve performance (requires v4.5+).
-        if not self._safe_mode and self.mercurial_version >= LooseVersion("4.5"):
+        if not self._safe_mode and self.mercurial_version >= Version("4.5"):
             self._config_options["rebase.experimental.inmemory"] = "true"
 
         # Enable evolve if the user's currently using it.  evolve makes amending
@@ -363,6 +372,20 @@ class Mercurial(Repository):
             self.obsstore = os.path.join(self.path, ".hg", "store", "obsstore")
             self.unlink_obsstore = not os.path.exists(self.obsstore)
 
+            # Warn users that `evolve` is not enabled and direct
+            # them to `mach vcs-setup`.
+            logger.warning(
+                "moz-phab detected that hg isn't configured with `evolve`. This "
+                "usually causes severe performance issues. Run `./mach vcs-setup` "
+                "to configure evolve."
+            )
+
+        # User may wish to use topics rather than bookmarks.
+        ext_topic = self._get_extension("topic", hg_config)
+        if ext_topic is not None:
+            self._safe_config_options["extensions.topic"] = ext_topic
+            self.use_topic = True
+
         # This script interacts poorly with mq.
         ext_mq = self._get_extension("mq", hg_config)
         self.has_mq = ext_mq is not None
@@ -517,7 +540,7 @@ class Mercurial(Repository):
 
         Args:
             node - SHA1 or revision of the base commit
-            name - name of the bookmark to be created
+            name - name of the bookmark/topic to be created
         """
         # Checkout sha
         if node:
@@ -538,6 +561,16 @@ class Mercurial(Repository):
             if not self.args.raw:
                 logger.info("Bookmark set to %s", bookmark_name)
 
+        if name and self.use_topic and config.create_topic and not self.args.no_topic:
+            topics = self.hg_out(["topics", "-T", "{topic}\n"])
+            topic_name = name
+            i = 0
+            while topic_name in topics:
+                i += 1
+                topic_name = "%s_%s" % (name, i)
+
+            self.hg(["topic", topic_name])
+
     def apply_patch(self, diff, body, author, author_date):
         changeset_str = self.format_patch(diff, body, author, author_date)
         with temporary_binary_file(changeset_str.encode("utf8")) as changeset_file:
@@ -676,6 +709,26 @@ class Mercurial(Repository):
         for node in non_stack_children:
             self.hg(["rebase", "--source", node, "--dest", commit["node"]])
 
+    def is_descendant(self, node: str) -> bool:
+        # Query the log for all commits that are both in the revset and descendants of
+        # `node`. If we get any results, then our revset is already based off of `node`.
+        out = self.hg_out(
+            ["log", "-r", f"{node}:: and {self.revset}", "-T", "{node}\n"],
+            split=False,
+        )
+
+        # If the command output is empty, our revset does not descend from `node`.
+        if not out:
+            return False
+
+        return True
+
+    def map_callsign_to_unified_head(self, callsign: str) -> Optional[str]:
+        if not self.is_node(callsign):
+            return None
+
+        return callsign
+
     def rebase_commit(self, source_commit, dest_commit):
         self.hg(
             ["rebase"]
@@ -683,10 +736,56 @@ class Mercurial(Repository):
             + ["--dest", dest_commit["node"]]
         )
 
+    def uplift_commits(self, dest: str, commits: List[dict]) -> List[dict]:
+        out = self.hg_out(
+            [
+                # Send messages to `stderr` so `stdout` is pure JSON.
+                "--config",
+                "ui.message-output=stderr",
+                "rebase",
+                "-r",
+                self.revset,
+                "-d",
+                dest,
+                # Don't obsolete the original changesets.
+                "--keep",
+                "-T",
+                "json",
+            ],
+            split=False,
+        )
+
+        # Capture the JSON output from rebase and use it to refresh our commit stack.
+        # We need to do this since using `--keep` with rebase causes the newly produced
+        # changesets to not be marked as successors of the original revisions, which
+        # means `refresh_commit_stack` won't find any updates to make.
+        nodechanges = json.loads(out)[0]["nodechanges"]
+
+        if len(nodechanges) != len(commits):
+            raise ValueError(
+                f"Rebase created {len(nodechanges)} new changesets; "
+                f"{len(commits)} expected"
+            )
+
+        for commit in commits:
+            # We should have a single new commit for each commit we want to uplift.
+            # Error out if we don't as progressing with an inconsistent stack might
+            # cause behaviour we can't handle.
+            new = nodechanges.get(commit["node"])
+            if len(new) != 1:
+                raise ValueError(f"Should only have a single new item, got {new}")
+
+            self._refresh_commit(commit, new[0])
+
+        # Set revset to the new range of commits.
+        self.revset = "%s::%s" % (commits[0]["node"], commits[-1]["node"])
+
+        return commits
+
     def check_commits_for_submit(self, commits, require_bug=True):
         # 'Greatest Common Ancestor'/'Merge Base' should be included in the revset.
         ancestor = self.hg_log("ancestor(%s)" % self.revset, split=False)
-        if ancestor not in [c["node"] for c in commits]:
+        if not any(commit["node"] == ancestor for commit in commits):
             raise Error(
                 "Non-linear commit stack (common ancestor %s missing from stack)"
                 % short_node(ancestor)
@@ -916,7 +1015,7 @@ class Mercurial(Repository):
     def _change_add(self, change, fn, _old_fn, _parent, node):
         """Create a change about adding a file to the commit."""
         meta = self._get_file_meta(fn, node)
-        telemetry.metrics.mozphab.submission.files_size.accumulate(meta["file_size"])
+        telemetry().submission.files_size.accumulate(meta["file_size"])
 
         if meta["binary"]:
             change.set_as_binary(
@@ -927,7 +1026,7 @@ class Mercurial(Repository):
             )
             return
 
-        # emtpy files don't have any hunks
+        # Empty files don't have any hunks.
         if meta["file_size"] == 0:
             return
 
@@ -949,7 +1048,7 @@ class Mercurial(Repository):
     def _change_del(self, change, fn, _old_fn, parent, _node):
         """Create a change about deleting a file from the commit."""
         meta = self._get_file_meta(fn, parent)
-        telemetry.metrics.mozphab.submission.files_size.accumulate(meta["file_size"])
+        telemetry().submission.files_size.accumulate(meta["file_size"])
 
         if meta["binary"]:
             change.set_as_binary(
@@ -960,7 +1059,7 @@ class Mercurial(Repository):
             )
             return
 
-        # emtpy files don't have any hunks
+        # Empty files don't have any hunks.
         if meta["file_size"] == 0:
             return
 
@@ -982,7 +1081,7 @@ class Mercurial(Repository):
         a_meta = self._get_file_meta(old_fn, parent)
         b_meta = self._get_file_meta(fn, node)
         file_size = max(a_meta["file_size"], b_meta["file_size"])
-        telemetry.metrics.mozphab.submission.files_size.accumulate(file_size)
+        telemetry().submission.files_size.accumulate(file_size)
 
         if a_meta["binary"] or b_meta["binary"]:
             # Binary file.
@@ -994,6 +1093,10 @@ class Mercurial(Repository):
             )
             return
 
+        # Empty files don't have any hunks.
+        if a_meta["file_size"] == 0 and b_meta["file_size"] == 0:
+            return
+
         if a_meta["body"] == b_meta["body"]:
             # File contents unchanged.
             lines = create_hunk_lines(a_meta["body"], " ", check_eof=False)[0]
@@ -1028,3 +1131,20 @@ class Mercurial(Repository):
                 file_size,
                 time.process_time() - start,
             )
+
+    def validate_email(self):
+        email = self.extract_email_from_username()
+        if not is_valid_email(email):
+            raise Error(
+                f"Your username configured with Mercurial ({email}) "
+                f"must contain a valid email.\n"
+                f"Please see https://www.mercurial-scm.org/doc/hgrc.5.html "
+                f"for more information on editing your Mercurial configuration."
+                f"\n\nYou can also amend a commit via "
+                f'`hg commit --amend --user "username <some@email.com>"'
+            )
+
+    def extract_email_from_username(self) -> str:
+        """Extracts an email from a Mercurial username, if it exists.
+        Not guaranteed to return a valid email, make sure to validate."""
+        return self.username.split("<").pop().replace(">", "")
