@@ -18,6 +18,11 @@ from .logger import logger
 
 USER_INFO_FILE = Path(environment.MOZBUILD_PATH) / "user_info.json"
 EMPLOYEE_CHECK_FREQUENCY = 24 * 7 * 60 * 60  # week
+MOZILLA_EMPLOYEE_EMAIL_ENDINGS = {
+    "@getpocket.com",
+    "@mozilla.com",
+    "@mozillafoundation.org",
+}
 
 
 def is_bad_uuid(key: str, value: Optional[str]) -> bool:
@@ -54,9 +59,8 @@ class UserData:
 
     @property
     def is_data_collected(self):
-        """True if all user info data are collected."""
-        # All values are set (not None)
-        return None not in [getattr(self, k) for k in self.keys]
+        """True if all user info data is collected."""
+        return all(getattr(self, k) is not None for k in self.keys)
 
     def to_dict(self):
         return {k: getattr(self, k) for k in self.keys}
@@ -74,10 +78,12 @@ class UserData:
 
     def set_from_file(self):
         """Read user info from file."""
-        if USER_INFO_FILE.exists():
-            with USER_INFO_FILE.open("r", encoding="utf-8") as f:
-                user_info = json.load(f)
-                self.update_from_dict(user_info)
+        if not USER_INFO_FILE.exists():
+            return
+
+        with USER_INFO_FILE.open("r", encoding="utf-8") as f:
+            user_info = json.load(f)
+            self.update_from_dict(user_info)
 
     def save_user_info(self, **kwargs):
         """Save any fields provided as kwargs into the user_info file."""
@@ -106,7 +112,10 @@ class UserData:
             response["is_employee"] = False
             return response
 
-        if response["email"].lower().endswith("@mozilla.com"):
+        lower_email = response["email"].lower()
+        if any(
+            lower_email.endswith(domain) for domain in MOZILLA_EMPLOYEE_EMAIL_ENDINGS
+        ):
             response["is_employee"] = True
             return response
 
